from pyinfra.api import FactBase


class FileContext(FactBase):
    '''
    Returns structured SELinux file context data for a specified file.

    .. code:: python
        {
            'user': 'system_u',
            'role': 'object_r',
            'type': 'deafult_t',
            'level': 's0',
        }
    '''

    def command(self, path):
        return 'stat -c %C {0} || exit 0'.format(path)

    def process(self, output):
        context = {}
        components = output[0].split(':')
        context['user'] = components[0]
        context['role'] = components[1]
        context['type'] = components[2]
        context['level'] = components[3]
        return context


class SEBoolean(FactBase):
    '''
    Returns the on/off status of a SELinux Boolean.

    .. code:: python
        host.get_fact(SEBoolean, "httpd_can_network_connect") -> "off"
    '''
    requires_command = 'getsebool'

    def command(self, boolean):
        return 'getsebool {0}'.format(boolean)

    def process(self, output):
        components = output[0].split(' --> ')
        return components[1]



import re

from pyinfra.api import FactBase


class OpenrcStatus(FactBase):
    '''
    Returns a dict of name -> status for OpenRC services for a given runlevel.
    '''

    default = dict
    requires_command = 'rc-status'
    regex = r'\s+([a-zA-Z0-9\-_]+)\s+\[\s+([a-z]+)\s+\]'

    def command(self, runlevel='default'):
        return 'rc-status {0}'.format(runlevel)

    def process(self, output):
        services = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                services[matches.group(1)] = matches.group(2) == 'started'

        return services


class OpenrcEnabled(FactBase):
    '''
    Returns a dict of name -> whether enabled for OpenRC services for a given runlevel.
    '''

    default = dict
    requires_command = 'rc-update'

    def command(self, runlevel='default'):
        self.runlevel = runlevel
        return 'rc-update show -v'

    def process(self, output):
        services = {}

        for line in output:
            name, levels = line.split('|', 1)
            name = name.strip()
            levels = levels.split()
            services[name] = self.runlevel in levels

        return services



from pyinfra.api import FactBase

from .util.packaging import parse_packages

PKGIN_REGEX = r'^([a-zA-Z\-0-9]+)-([0-9\.]+\-?[a-z0-9]*)\s'


class PkginPackages(FactBase):
    '''
    Returns a dict of installed pkgin packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'pkgin list'
    requires_command = 'pkgin'

    default = dict

    def process(self, output):
        return parse_packages(PKGIN_REGEX, output, lower=False)



import six


config_defaults = {
    # % of hosts which have to fail for all operations to stop
    'FAIL_PERCENT': None,

    # Seconds to timeout SSH connections
    'CONNECT_TIMEOUT': 10,

    # Temporary directory (on the remote side) to use for caching any files/downloads
    'TEMP_DIR': '/tmp',

    # Gevent pool size (defaults to #of target hosts)
    'PARALLEL': None,

    # Specify the required pyinfra version (using PEP 440 setuptools specifier)
    'REQUIRE_PYINFRA_VERSION': None,
    # Specify any required packages (either using PEP 440 or a requirements file)
    # Note: this can also include pyinfra, potentially replacing REQUIRE_PYINFRA_VERSION
    'REQUIRE_PACKAGES': None,

    # COMPAT w/<1.1
    # TODO: remove this in favour of above at v2
    # Specify a minimum required pyinfra version for a deploy
    'MIN_PYINFRA_VERSION': None,

    # All these can be overridden inside individual operation calls:

    # Switch to this user (from ssh_user) using su before executing operations
    'SU_USER': None,
    'USE_SU_LOGIN': False,
    'SU_SHELL': None,
    'PRESERVE_SU_ENV': False,

    # Use sudo and optional user
    'SUDO': False,
    'SUDO_USER': None,
    'PRESERVE_SUDO_ENV': False,
    'USE_SUDO_LOGIN': False,
    'USE_SUDO_PASSWORD': False,

    # Use doas and optional user
    'DOAS': False,
    'DOAS_USER': None,

    # Use doas and optional user
    'DOAS': False,
    'DOAS_USER': None,

    # Only show errors, but don't count as failure
    'IGNORE_ERRORS': False,

    # Shell to use to execute commands
    'SHELL': 'sh',
}


class Config(object):
    '''
    The default/base configuration options for a pyinfra deploy.
    '''

    state = None

    def __init__(self, **kwargs):
        # Always apply some env
        env = kwargs.pop('ENV', {})
        self.ENV = env

        config = config_defaults.copy()
        config.update(kwargs)

        for key, value in six.iteritems(config):
            setattr(self, key, value)

    def get_current_state(self):
        return [
            (key, getattr(self, key))
            for key in config_defaults.keys()
        ]

    def set_current_state(self, config_state):
        for key, value in config_state:
            setattr(self, key, value)

    def lock_current_sate(self):
        self._locked_config = self.get_current_state()

    def reset_locked_state(self):
        self.set_current_state(self._locked_config)



'''
Manage pkgin packages.
'''

from pyinfra.api import operation
from pyinfra.facts.pkgin import PkginPackages

from .util.packaging import ensure_packages


@operation
def upgrade(state=None, host=None):
    '''
    Upgrades all pkgin packages.
    '''

    yield 'pkgin -y upgrade'

_upgrade = upgrade  # noqa: E305


@operation
def update(state=None, host=None):
    '''
    Updates pkgin repositories.
    '''

    yield 'pkgin -y update'

_update = update  # noqa: E305


@operation
def packages(
    packages=None, present=True, latest=False,
    update=False, upgrade=False,
    state=None, host=None,
):
    '''
    Add/remove/update pkgin packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``pkgin update`` before installing packages
    + upgrade: run ``pkgin upgrade`` before installing packages

    Examples:

    .. code:: python

        # Update package list and install packages
        pkgin.packages(
            name='Install tmux and Vim',
            packages=['tmux', 'vim'],
            update=True,
        )

        # Install the latest versions of packages (always check)
        pkgin.packages(
            name='Install latest Vim',
            packages=['vim'],
            latest=True,
        )
    '''

    if update:
        yield _update(state=state, host=host)

    if upgrade:
        yield _upgrade(state=state, host=host)

    # TODO support glob for specific versions (it isn't as simple
    # as apt's, as pkgin supports something like 'mysql-server>=5.6<5.7')
    yield ensure_packages(
        host, packages, host.get_fact(PkginPackages), present,
        install_command='pkgin -y install',
        uninstall_command='pkgin -y remove',
        upgrade_command='pkgin -y upgrade',
        latest=latest,
    )



from unittest import TestCase

from pyinfra.api import Inventory


class TestInventoryApi(TestCase):
    def test_create_inventory(self):
        inventory = Inventory((['somehost', 'anotherhost', 'anotherotherhost'], {}))
        assert len(inventory) == 3

    def test_create_inventory_data(self):
        default_data = {'default_data': 'default_data'}
        inventory = Inventory((['somehost'], default_data))
        assert inventory.get_data() == default_data
        assert inventory.get_host_data('somehost') == {}

    def test_create_inventory_host_data(self):
        default_data = {'default_data': 'default_data', 'host_data': 'none'}
        somehost_data = {'host_data': 'host_data'}

        inventory = Inventory((
            [('somehost', somehost_data), 'anotherhost'],
            default_data,
        ))

        assert inventory.get_data() == default_data
        assert inventory.get_host_data('somehost') == somehost_data
        assert inventory.get_host_data('anotherhost') == {}
        assert inventory.get_host('anotherhost').data.host_data == 'none'

    def test_create_inventory_override_data(self):
        default_data = {'default_data': 'default_data', 'override_data': 'ignored'}
        override_data = {'override_data': 'override_data'}
        somehost_data = {'host_data': 'host_data'}

        inventory = Inventory((
            [('somehost', somehost_data), 'anotherhost'],
            default_data,
        ), override_data=override_data)

        assert inventory.get_data() == default_data
        assert inventory.get_override_data() == override_data

        assert inventory.get_host('somehost').data.host_data == 'host_data'
        assert inventory.get_host('anotherhost').data.host_data is None

        assert inventory.get_host('somehost').data.override_data == 'override_data'
        assert inventory.get_host('anotherhost').data.override_data == 'override_data'
